def __map_extensions__():
    from pathlib import Path
    import os
    from importlib import import_module
    from inspect import isclass
    from lolpop.component.base_component import BaseComponent
    from lolpop.pipeline.base_pipeline import BasePipeline
    from lolpop.runner.base_runner import BaseRunner
    import warnings

    warnings.filterwarnings("ignore")
    #get current directory and all subdirectors. These represent the extensions 
    # resource types
    path = Path(__file__).parent.resolve() #lolpop/extension
    extensions = ["%s/%s" % (path, x) for x in os.listdir(path)
                  if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
    #iterate through extensions and add each top level 
    for extension in extensions: #lolpop/extension/my_extension
        resources = ["%s/%s" % (extension, x) for x in os.listdir(extension)
                     if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
        for resource in resources: #lolpop/extension/my_extension/{component,pipeline,runner}
            subdirs = ["%s/%s" % (resource, x) for x in os.listdir(resource)
                       if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
        #for each resource type (i.e. subdir), get all resources implemented in that type (i.e. python files in the subdir)
        for subdir in subdirs:
            files = [x for x in os.listdir(subdir) if (
                (".py" in x) and ("__" not in x) and (x[0] != "."))]
            #from each file, import all classes and register them in the global namespace.
            for file in files:
                subdir_arr = subdir.split("/")
                module = import_module("lolpop.extension.%s.%s.%s.%s" % (
                    subdir_arr[-3], subdir_arr[-2], subdir_arr[-1], file[:-3]))
                classes = [x for x in dir(module) if isclass(getattr(module, x))]
                components = [x for x in classes if (
                    issubclass(getattr(module, x), BaseComponent)
                    or issubclass(getattr(module, x), BasePipeline)
                    or issubclass(getattr(module, x), BaseRunner)
                    )]
                globals().update({name: getattr(module, name)
                                for name in components})

        warnings.resetwarnings()


__map_extensions__()



from lolpop.component.base_component import BaseComponent
from lolpop.utils import common_utils as utils
import numpy as np 

class BaseModelChecker(BaseComponent): 

    def check_model(self, data, model, **kwargs): 
        pass 


    def get_baseline_comparison(self, data, model, model_version):

        #TODO: we are already getting predictions elsewhere. 
        #We should just save them and pass in instead of calculating again
        baseline_method = self._get_config("baseline_method")
        baseline_value = self._get_config("baseline_value")
        test_predictions = model.predict(data)
        baseline_predictions = self._get_baseline_predictions(data, baseline_method, baseline_value)

        #TODO: model metrics are being recalculated here, should save and load
        perf_metric = self._get_config("perf_metric")
        model_metrics = model.calculate_metrics(data, test_predictions, [perf_metric])
        baseline_metrics = model.calculate_metrics(data, baseline_predictions, [perf_metric])
        for split in baseline_metrics.keys(): 
            self.metrics_tracker.log_metric(model_version, "baseline_%s_%s" %(split, perf_metric), baseline_metrics[split][perf_metric])
        is_baseline_better, metric_diff = self._get_metric_comparison(model_metrics, baseline_metrics, perf_metric)

        self.metrics_tracker.log_metric(model_version, "model_test_performance_over_baseline", metric_diff)

        if is_baseline_better: 
            self.notify("Model performed worse than baseline by %s%%" %metric_diff, level="WARN")

        return metric_diff 

    def compare_models(self, data, model, prev_model, model_version, prev_model_version=None): 
        #this should only happen during the first training of a model
        if prev_model_version == model_version:
            is_new_model_better = True 
        else: 
            perf_metric = self._get_config("perf_metric")
            is_new_model_better, metric_diff = self._compare_model_performance(model, prev_model, data, perf_metric, model_version)

        return is_new_model_better

    #default implementation of baseline predictions
    def _get_baseline_predictions(self, data, baseline_method, baseline_value): 
        
        has_test = "X_test" in data.keys()
        baseline_predictions = {}

        #column = user defined column should be used for baseline comparison
        if baseline_method == "column": 
            baseline_predictions["train"] = data["X_train"][baseline_value]
            baseline_predictions["valid"] = data["X_valid"][baseline_value]
            if has_test: 
                baseline_predictions["test"] = data["X_test"][baseline_value]

        elif baseline_method == "value":    
            labels = data["y_train"]
            #this is kind of hacky but ... shrug
            labels.index=labels.values
            label_counts = labels.groupby(level=0).count()

            if baseline_value == "class_avg": 
                weights = [x/len(labels) for x in label_counts]
                baseline_predictions["train"] = np.random.choice(label_counts.index, p=weights, size=len(data["X_train"]))
                baseline_predictions["valid"] = np.random.choice(label_counts.index, p=weights, size=len(data["X_valid"]))
                if has_test: 
                    baseline_predictions["test"] = np.random.choice(label_counts.index, p=weights, size=len(data["X_test"]))

            elif baseline_value == "class_most_frequent": 
                most_frequent_class = sorted(label_counts.items(), key=lambda x: x[1], reverse=True)[0][0] #get most frequent class 
                baseline_predictions["train"] = [most_frequent_class] * len(data["X_train"])
                baseline_predictions["valid"] = [most_frequent_class] * len(data["X_valid"])
                if has_test: 
                    baseline_predictions["test"] = [most_frequent_class] * len(data["X_test"])

        else: #consider other ways to do baseline comparison
            pass

        return baseline_predictions 

    def _get_metric_comparison(self, champion_metric, challenger_metric, perf_metric): 
        old_metric = champion_metric.get("test").get(perf_metric)
        new_metric = challenger_metric.get("test").get(perf_metric)
        lower_is_better = utils.get_metric_direction(perf_metric)

        if lower_is_better: #True = perf_metric is better if lower
            is_challenger_better = (new_metric < old_metric) 
            metric_diff = (new_metric - old_metric) / max(old_metric, 0.00001) * 100
        else: #wants higher perf metric
            is_challenger_better = (new_metric > old_metric)
            metric_diff = (old_metric - new_metric) / max(old_metric, 0.00001) * 100 

        return is_challenger_better, metric_diff

    def _compare_model_performance(self, model, deployed_model, data, perf_metric, current_model_version): 
        current_predictions = model.predict(data)
        deployed_predictions = deployed_model.predict(data)

        current_metrics = model.calculate_metrics(data, current_predictions, [perf_metric])
        deployed_metrics = deployed_model.calculate_metrics(data, deployed_predictions, [perf_metric])

        is_new_model_better, metric_diff = self._get_metric_comparison(deployed_metrics, current_metrics, perf_metric)

        self.metrics_tracker.log_metric(current_model_version, "deployed_model_perf_metric_diff", metric_diff)

        return is_new_model_better, metric_diff

    def calculate_model_drift(self, data, model, deployed_model, **kwargs):
        pass


from lolpop.pipeline.train.base_train import BaseTrain
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class OfflineTrain(BaseTrain): 
    __REQUIRED_CONF__ = {
        "components": ["data_splitter", "metadata_tracker", "model_checker", "model_explainer", "model_visualizer", "model_bias_checker"], 
        "config": []
    }

    def split_data(self, data, **kwargs): 
        #split data. returns dictionary of train/vali/test dataframes
        data_dict = self.data_splitter.split_data(data)
        
        return data_dict

    def train_model(self, data, *args, **kwargs): 
        #create model_version
        id = self._get_config("model_name")
        model_version = self.metadata_tracker.create_resource(id, type="model_version")
    
        #if we are using hyperparameter tuner, use that, otherwise just use the model trainer provided
        if hasattr(self, "hyperparameter_tuner"): 
            model = self.hyperparameter_tuner.run_experiment(data, model_version)
        else: 
            #TODO: this needs a better entry point. build_model doesn't actually log stuff
            model = self.model_trainer.build_model(data, model_version)

        return model, model_version

    def check_model(self, data_dict, model, model_version, **kwargs):
        #run model checks
        model_report, file_path, checks_status = self.model_checker.check_model(data_dict, model)

        #log model report to metadata tracker
        self.metadata_tracker.log_checks(
            model_version,
            file_path = file_path, 
            report = model_report, 
            checker_class = type(self.model_checker).__name__, 
            type="model"
            )

        if checks_status == "ERROR" or checks_status == "WARN": 
            url = self.metadata_tracker.url
            self.notify("Issues found with model checks. Visit %s for more information." %url, checks_status)


    def analyze_model(self, data_dict, model, model_version, **kwargs): 
        #calculate feature importance
        self.model_explainer.get_feature_importance(data_dict, model, model_version)
         
        #compare model to baseline
        self.model_checker.get_baseline_comparison(data_dict, model, model_version)

        #create some eye candy
        self.model_visualizer.generate_viz(data_dict, model._get_model(), model_version)


    def compare_models(self, data_dict, model, model_version): 
        #get currently deployed model version/previous model version and winning experiment
        prev_model_version = self.metadata_tracker.get_currently_deployed_model_version(model_version)

        if prev_model_version is not None: 
            prev_experiment = self.metadata_tracker.get_winning_experiment(prev_model_version)

            #get the prev model version object from resource version control and load it into a model trainer
            prev_model_obj = self.resource_version_control.get_model(prev_experiment)
            prev_model = self.metadata_tracker.load_model(prev_model_obj, model_version, model)

            #compare model to the last version 
            report, file_path = self.model_checker.calculate_model_drift(data_dict, model, prev_model)
            self.metadata_tracker.log_artifact(model_version, id="model_drift_report", path = file_path, external = False)
            is_new_model_better = self.model_checker.compare_models(data_dict, model, prev_model, model_version, prev_model_version)
            
            return is_new_model_better 
        
        else: 
            self.log("No previous model version found for model: %s" %self.metadata_tracker.get_resource_id(model_version))
            return True 

    def check_model_bias(self, data_dict, model, model_version):
        #check model bias
        self.model_bias_checker.check_model_bias(data_dict, model, model_version)

    def retrain_model_on_all_data(self, data, model_version, ref_model = None):
        if hasattr(self, "model_trainer"):  
            model, experiment = self.model_trainer.rebuild_model(data, model_version)
        else: 
            #if no model_trainer is specified in config, we can rebuild using the winning_exp_model_trainer
            winning_exp_model = self.metadata_tracker.load_model(None, model_version, ref_model=ref_model)
            model, experiment = winning_exp_model.rebuild_model(data, model_version)
            return model, experiment



from lolpop.component.model_bias_checker.base_model_bias_checker import BaseModelBiasChecker
from lolpop.utils import common_utils as utils
from aif360.datasets import StandardDataset
from aif360.metrics import BinaryLabelDatasetMetric, ClassificationMetric

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class AIFairnessModelBiasChecker(BaseModelBiasChecker): 

    def check_model_bias(self, data, model, model_version, *args, **kwargs): 
        #set up data/predictions
        df_train = data["X_train"]
        df_train_w_labels = df_train.copy() 
        df_train_w_labels[self._get_config("model_target")] = data["y_train"].values
        preds = model.predict(data).get("train")

        #set up aif dataset + configs
        model_target = self._get_config("model_target")
        privileged_groups = self._get_config("privileged_groups")
        unprivileged_groups = self._get_config("unprivileged_groups")
        ds = StandardDataset(
            df=df_train_w_labels.dropna(), #NA make AI360 fail 
            label_name=model_target, 
            protected_attribute_names=self._get_config("protected_attribute_names"), 
            favorable_classes=self._get_config("favorable_classes"), 
            privileged_classes=self._get_config("privileged_classes"),
            )

        #get fairness metrics from binarylabeldatasetmetric. This essentially checks bias in the dataset itself
        bldm = BinaryLabelDatasetMetric(ds, unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)
        self.metrics_tracker.log_metric(model_version, "bldm_base_rate", bldm.base_rate()) # ratio of priv/total 
        self.metrics_tracker.log_metric(model_version, "bldm_disparate_impact", bldm.disparate_impact()) # ratio of prob fav outcomes of non/priv
        self.metrics_tracker.log_metric(model_version, "bldm_mean_difference",  bldm.mean_difference()) # diff in probability of favorable outcoems for non vs priv
        self.metrics_tracker.log_metric(model_version, "bldm_consistency", bldm.consistency()[0]) # meastures how similar the labels are for similar instanes
        self.metrics_tracker.log_metric(model_version, "bldm_smoothed_edf", bldm.smoothed_empirical_differential_fairness()) #

        ds_preds = ds.copy()
        ds_preds.labels=preds

        #get fairness metric from classificationmetric. This checks fairness in predictions
        cm = ClassificationMetric(ds, ds_preds, unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)
        self.metrics_tracker.log_metric(model_version, "cm_tpr_diff", cm.true_positive_rate_difference()) 
        self.metrics_tracker.log_metric(model_version, "cm_selection_rate", cm.selection_rate()) 
        self.metrics_tracker.log_metric(model_version, "cm_theil_index", cm.theil_index()) 
        self.metrics_tracker.log_metric(model_version, "cm_false_positive_rate_ratio", cm.false_positive_rate_ratio()) 
        self.metrics_tracker.log_metric(model_version, "cm_false_omission_rate_ratio", cm.false_omission_rate_ratio()) 
        self.metrics_tracker.log_metric(model_version, "cm_false_negative_rate_ratio", cm.false_negative_rate_ratio())
        self.metrics_tracker.log_metric(model_version, "cm_false_discovery_rate_ratio", cm.false_discovery_rate_ratio()) 
        self.metrics_tracker.log_metric(model_version, "cm_error_rate_ratio", cm.error_rate_ratio())  
        self.metrics_tracker.log_metric(model_version, "cm_differential_fba", cm.differential_fairness_bias_amplification())
        
        #TODO: we should have bias thresholds, and if any are exceeded, then we can reweigh the dataset
        #via something like: https://aif360.readthedocs.io/en/stable/modules/generated/aif360.algorithms.preprocessing.Reweighing.html
        #will have to think about how this applies in the workflow though.
        #i.e. we may want to move the dataset bias stuff up in the data_processing step. 


from lolpop.component.base_component import BaseComponent
from sklearn import metrics as sk_metrics
import pandas as pd 

class BaseModelTrainer(BaseComponent): 

    model = None 
    mlflow_module = "you_need_to_implement_this_for_mlflow_use"
    
    #predictions = {}
    def __init__(self, params=None, *args, **kwargs): 
        #set normal config
        super().__init__(params=params, *args, **kwargs)
        self.params = params

    def fit(self, data, *args, **kwargs): 
        pass 

    def predict(self, data, *args, **kwargs): 
        pass
    
    def save(self, experiment, *args, **kwargs): 
        algo = type(self).__name__
        vc_info = self.resource_version_control.version_model(experiment, self.model, algo=algo)
        experiment_metadata = {
            "training_params" : self.params,
            "model_trainer" : algo
        }
        self.metadata_tracker.register_vc_resource(experiment, vc_info, additional_metadata = experiment_metadata)


    def _predict_df(self, df):
        pass 

    def _predict_proba_df(self, df): 
        pass 

    def _get_model(self):
        return self.model 
    
    #used to set model object. 
    def _set_model(self, model): 
        self.model = model 
    #def _get_predictions(self): 
    #    return self.predictions

    def calculate_metrics(self, data, predictions, metrics, **kwargs): 
        metrics_out = {"train" : {}, "valid" : {}, "test" : {}}

        test_exists = data.get("y_test") is not None
        valid_exists = data.get("y_valid") is not None

        if self.problem_type == "classification": 
            multi_class = False
            average="binary"
            num_classes = len(data["y_train"].unique())
            if num_classes > 2: 
                multi_class = True 
                average = "weighted"

        for metric in metrics: 
            if metric == "accuracy": 
                metrics_out["train"][metric] = sk_metrics.accuracy_score(data["y_train"], predictions["train"])
                if valid_exists: 
                    metrics_out["valid"][metric] = sk_metrics.accuracy_score(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.accuracy_score(data["y_test"], predictions["test"])
            if metric == "f1": 
                metrics_out["train"][metric] = sk_metrics.f1_score(data["y_train"], predictions["train"], average = average)
                if valid_exists: 
                    metrics_out["valid"][metric] = sk_metrics.f1_score(data["y_valid"], predictions["valid"], average = average)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.f1_score(data["y_test"], predictions["test"], average = average)
            if metric == "rocauc": 
                metrics_out["train"][metric] = sk_metrics.roc_auc_score(data["y_train"], predictions["train_proba"], average = average, multi_class = "ovr")
                if valid_exists: 
                    metrics_out["valid"][metric] = sk_metrics.roc_auc_score(data["y_valid"], predictions["valid_proba"], average = average, multi_class = "ovr")
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.roc_auc_score(data["y_test"], predictions["test_proba"], average = average, multi_class = "ovr")
            if metric == "prauc": 
                if not multi_class: #only works for binary
                    metrics_out["train"][metric] = sk_metrics.average_precision_score(data["y_train"], predictions["train_proba"][:,1])
                    if valid_exists: 
                        metrics_out["valid"][metric] = sk_metrics.average_precision_score(data["y_valid"], predictions["valid_proba"][:,1])
                    if test_exists:
                        metrics_out["test"][metric] = sk_metrics.average_precision_score(data["y_test"], predictions["test_proba"][:,1])
            if metric == "precision": 
                metrics_out["train"][metric] = sk_metrics.precision_score(data["y_train"], predictions["train"], average = average)
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.precision_score(data["y_valid"], predictions["valid"], average = average)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.precision_score(data["y_test"], predictions["test"], average = average)
            if metric == "recall": 
                metrics_out["train"][metric] = sk_metrics.recall_score(data["y_train"], predictions["train"], average = average)
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.recall_score(data["y_valid"], predictions["valid"], average = average)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.recall_score(data["y_test"], predictions["test"], average = average)

        return metrics_out

    def build_model(self, data, model_version, algo, params): 
        #fit model
        model_obj = self.fit(data)

        #create experiment and save model 
        experiment = self.metadata_tracker.create_resource(id=None, type="experiment", parent=model_version)
        self.save(experiment)

        #get predictions
        predictions = self.predict(data)

        #calculate metrics
        metrics_val = self.calculate_metrics(data, predictions, self._get_config("metrics"))

        #log stuff
        self.metrics_tracker.log_metrics(experiment, metrics_val, self._get_config("perf_metric"))
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_id", data={"winning_experiment_id" : self.metadata_tracker.get_resource_id(experiment)})
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_model_trainer", data={"winning_experiment_model_trainer" : type(self).__name__})

        #save splits
        for k,v in data.items(): 
            vc_info = self.resource_version_control.version_data(model_version, v, key=k)
            self.metadata_tracker.register_vc_resource(model_version, vc_info, key=k, file_type="csv")

        return self, experiment 

    def rebuild_model(self, data, model_version):
        has_test = ("X_test" in data.keys())

        #merge all dfs 
        df_X = pd.concat([data["X_train"], data["X_valid"]])
        df_y = pd.concat([data["y_train"], data["y_valid"]])
        if has_test: 
            df_X = pd.concat([df_X, data["X_test"]])
            df_y = pd.concat([df_y, data["y_test"]])
        train_data = {"X_train" : df_X, "y_train": df_y}

        #get params, class from winning experiment 
        winning_experiment = self.metadata_tracker.get_winning_experiment(model_version)
        params = self.metadata_tracker.get_metadata(winning_experiment, "training_params")
        model_trainer = self.metadata_tracker.get_metadata(winning_experiment, "model_trainer")
        
        #rebuild model on all data 
        #we need to use the single model entry point for this, as this doesn't log the rigth info
        model, exp = self.build_model(train_data, model_version, model_trainer, params)                        
        
        return model, exp


import subprocess 
import os 
import sys
from importlib import import_module, util as import_util 
from git import Repo
from omegaconf import OmegaConf, dictconfig
from datetime import datetime
from functools import wraps 
import pandas as pd 
from pathlib import Path 
import time

#
##should decide if these things go into the common_utils class or just the abstract classes. 
#

#shell command wrapper
def execute_cmd(cmd, logger=None):
    if logger: 
        logger.log("Executing command: `%s`" %(" ".join(cmd)))
    p = subprocess.run(cmd, stdout=subprocess.PIPE, universal_newlines=True, shell=False)
    return (p.stdout, p.returncode)

# pulls keys out of config or gets from env variable
# i.e. either handle secret manager before getting here and pass in 
# or populate via env vars
def load_config(config_keys, conf): 
    config = {}
    for key in config_keys: 
        if key in conf.keys():
            config[key]=conf[key]
        elif key.lower() in conf.keys(): #handle cases where we're looking for env var but it's overridden in conf w/ lowercase 
            config[key] = conf[key.lower()]
        else: 
            config[key] = os.getenv(key)
    return config

#comits a single file to git
def git_commit_file(file_path, repo_path=None, msg="Commiting file from lolpop", push=True, logger=None):
    repo = Repo(repo_path)
    repo.index.add(file_path)
    hexsha = repo.index.commit(msg).hexsha

    if push: 
        origin = repo.remotes[0]
        result = origin.push()
        if logger: 
            logger.log("Committed and pushed file %s. Result: %s" %(file_path, result[0].summary))

    return hexsha

def load_plugins(plugin_paths=[]):
    plugins = []
    for dir in plugin_paths:
        if dir.exists():
            plugin = load_plugin(dir)
            plugins.append(plugin)
    return plugins

def load_plugin(plugin_path, obj=None): 
    if plugin_path.is_dir(): 
        plugin_dir = str(plugin_path)
        plugin_name = plugin_path.name
    elif plugin_path.is_file():
        plugin_dir = str(plugin_path.parent)
        plugin_name = plugin_path.stem
    else: 
        raise Exception("Invalid plugin path: %s. Path is not a file nor a directory." %plugin_path)
    if plugin_dir: 
        sys.path.append(plugin_dir)
    if plugin_name: 
        mod = import_module(plugin_name)
    if obj: 
        plugin_paths = obj._get_config("plugin_paths",[]) 
        plugin_paths.append(plugin_dir)
        obj._set_config("plugin_paths", plugin_paths)
    return mod

#load class object
def load_class(class_name, class_type="component", parent="lolpop"): 
    module = import_module("%s.%s" %(parent, class_type))
    cl = getattr(module, class_name)
    return cl

#load class object from plugins
def load_class_from_plugin(class_name, plugin_mods, class_type="component"):
    #try to load from each plugin
    cl = None 
    for plugin in plugin_mods: 
        try: 
            cl = load_class(class_name, class_type=class_type, parent=plugin)
            break 
        except Exception as e: 
            #if multiple plugins are used, the one we are looking for will not be in most of them, so just ignore any errors
            continue 
    return cl

#loads a module from a file
def load_module_from_file(file_path): 
    spec = import_util.spec_from_file_location(file_path.stem, file_path)
    mod = import_util.module_from_spec(spec)
    sys.modules[spec.name]=mod #need to do this to properly register module
    spec.loader.exec_module(mod)
    return mod 

#register component class as an attribute of the provided object
def register_component_class(self_obj, conf, component_type, default_class_name=None, pipeline_conf = {}, runner_conf = {}, parent_process = "runner", problem_type = None, dependent_components = {}, plugin_mods=[]): 
    obj = None
    component_class_name = conf.components.get(component_type, default_class_name)
    if component_class_name is not None:
        obj = None
        try: 
            cl = load_class(component_class_name) 
        except: 
            self_obj.log(
            	"Unable to find class %s in built-in components. Searching plugins modules in %s..." %(component_class_name, str(plugin_mods)))
            cl = load_class_from_plugin(component_class_name, plugin_mods)
            if cl is not None: 
                self_obj.log("Found class %s in plugins!" % component_class_name)
            else: 
                self_obj.log("Unable to find class %s in plugins!" %component_class_name)
        if cl is not None: 
            obj = cl(conf=conf.get(component_type, {}), pipeline_conf=pipeline_conf, runner_conf=runner_conf,
                     parent_process=parent_process, problem_type=problem_type, components=dependent_components)
            setattr(self_obj, component_type, obj)
    return obj 

#registers pipeline as an attribute of the provided object
def register_pipeline_class(self_obj, conf, pipeline_type, default_class_name=None, runner_conf = {}, parent_process = "runner", problem_type = None, dependent_components = {}, plugin_mods=[]): 
    obj = None 
    pipeline_class_name = conf.pipelines.get(pipeline_type, default_class_name)
    if pipeline_class_name is not None: 
        try: 
            cl = load_class(pipeline_class_name, class_type="pipeline")
        except: 
            self_obj.log(
            	"Unable to find pipeline %s in built-in pipelines. Searching plugins..." % pipeline_class_name)
            cl = load_class_from_plugin(
            	pipeline_class_name, plugin_mods, class_type="pipeline")
            self_obj.log(
            	"Found class %s in plugins!" % pipeline_class_name)
        obj = None
        if cl is not None: 
            obj = cl(conf=conf.get(pipeline_type, {}), runner_conf=runner_conf, parent_process=parent_process,
                     problem_type=problem_type, components=dependent_components, plugin_mods=plugin_mods)
            setattr(self_obj, pipeline_type, obj)
    return obj

def lower_conf(conf): 
    return {k.lower():v for k,v in conf.items()}
    
#Need to revisit. Seems like this is not actually merging on update, it's reassigning. 
#As a result, some default conf can get lost
def copy_config_into(conf, default_conf): 
    #have to use omegaconf so types match
    default_copy = OmegaConf.create(default_conf.copy())
    updated_conf = OmegaConf.merge(default_copy, conf)
    return updated_conf

#validates configuration with the required conf specified
def validate_conf(config, required_conf, components_objs={}):
    conf = config.copy()
    #convert componenets_obj dict into primitives
    components = OmegaConf.create({k:type(v).__name__ for k,v in components_objs.items()})
    #updates conf.components with components. 
    #we want local config values to overwrite glboals, so we do a two step update
    if conf.get("components", None) is None: 
        conf["components"]=components
    else: 
        components.update(conf.get("components",{}))
        conf.get("components",{}).update(components)

    missing = {}
    total_missing = 0
    if required_conf is not None: 
        if conf is None: 
            conf = {}
        #map everything to lowercase keys to avoid capitalization mismatch
        lconf = lower_conf(conf)
        lrequired_conf = lower_conf(required_conf)
        for k,v in lrequired_conf.items(): 
            missing_k = []
            for val in v: 
                #value can be of form "key|values" which specifies the value(s) allowed
                #this is mainly used to specify the classes supported
                #i.e. something like "MetadataTracker|ContinualMetadataTracker"
                if "|" in val: 
                    val_arr = val.split("|")
                    val_key = val_arr[0]
                    val_value = val_arr[1].split(",")
                else: 
                    val_key = val
                    val_value = None
                if lconf.get(k,{}).get(val_key.lower(),None) is None: 
                    missing_k.append(val) 
                elif val_value is not None: 
                    if lconf.get(k,{}).get(val_key.lower()) not in val_value: 
                        missing_k.append(val)
            missing[k] = missing_k
        for v in missing.values(): 
            total_missing = total_missing + len(v)
    return missing, total_missing

#replaces variable values with their actual values. 
#conf allows pointing to other variable via $node1.node2.node3, etc...
def resolve_conf_variables(conf, main_conf=None):
    if main_conf is None: #need to always have the full conf lineage because variables can pont to arbitrary nodes
        main_conf = conf
    if conf is not None and len(conf) > 0: 
        for k,v in conf.items(): 
            if isinstance(v, dictconfig.DictConfig):
                OmegaConf.update(conf, k, resolve_conf_variables(v, main_conf)) 
            elif isinstance(v, str): 
                if v[0] == "$": 
                    resolved_value = get_conf_value(v[1:], main_conf)
                    if resolved_value: 
                        OmegaConf.update(conf,k, resolved_value)
                    else: 
                        raise Exception("Unable to resolve configuration value: %s" %v)
    return conf

# returns configuration from conf object (file location or python dict)
def get_conf(conf_obj): 
    if isinstance(conf_obj, str):
        conf = OmegaConf.load(conf_obj)
    elif type(conf_obj) is dict: 
        conf = OmegaConf.create(conf_obj)
    elif isinstance(conf_obj, dictconfig.DictConfig):
        conf = conf_obj
    else: 
        raise Exception("Invalid configuration. Configuration must be a file or a a dict.")
    
    return conf 

#returns node value in conf specified by var in the form 'node1.node2.node3...'
def get_conf_value(var, conf): 
    var_arr = var.split(".")
    out = conf
    try: 
        for i in var_arr: 
            out = out.get(i)
    except: 
        out = None 
    return out 

def get_plugin_mods(self_obj, plugin_paths=[], file_path=None):
    # if no plugin_dir is provided, then try to use the parent directory.
    # if the parent directory is lolpop, then it is a built-in runner and we can ignore
    if file_path is not None:
        #directory should be something like <module_name>/<runner>/<runner_type>/<runner_class>.py
        plugin_dir = os.path.dirname(os.path.dirname(
            os.path.dirname(os.path.realpath(file_path))))
        plugin_paths.append(plugin_dir)

    plugin_paths = [Path(dir) for dir in plugin_paths if dir != "lolpop"]

    #load up all plugin modules
    plugins = load_plugins(plugin_paths)
    plugin_mods = []
    for plugin in plugins: 
        if plugin is not None: 
            plugin_mods.append(plugin.__name__)
    
    return plugin_mods

#for a given lolpop object, get all plugin paths in children
def get_all_plugin_paths(obj):
    children = [x for x in dir(obj) if not x.startswith("_")]
    plugin_paths=[]

    for child in children: 
        child_obj = getattr(obj, child)
        if hasattr(child_obj, "_get_config"): 
            paths = child_obj._get_config("plugin_paths", None)
            if paths is not None: 
                plugin_paths = plugin_paths + paths
    
    plugin_paths = [Path(path) for path in set(plugin_paths)]

    return plugin_paths

#wraps logging calls around function execution
def log_execution(level="DEBUG", start_msg = None, end_msg = None, timeit=True):
    def log_decorator(func):
        @wraps(func)
        def wrapper(obj, *args, **kwargs):
            start = start_msg
            end = end_msg
            if not start: 
                start = "Starting execution of %s" %(func.__name__)
            if not end: 
                end = "Finished execution of %s" %(func.__name__)
            obj.log(start, level)
            obj.log("args: %s, kwargs: %s" %(args, kwargs), "TRACE")
            start_time = time.process_time()
            result = func(obj, *args, **kwargs)
            end_time = time.process_time()
            if timeit: 
                end = end + ". Completed in %s seconds." %(str(end_time - start_time))
            obj.log(end, level)
            return result
        return wrapper
    return log_decorator

#wraps function around error handling
def error_handler(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise Exception(f"An error occurred: {e}")
    return wrapper

# decorate all public methods in a class with decorator
# should only be applied to class definition
# decorators can be a single decorator or list of decorators that will be applied in reverse order
# i.e. the first decorator in the list would be the topmost decorator if listed out
# (this seems like the intuitive thing to do)
def decorate_all_methods(decorators):
    def decorate(cls):
        decorator_list = decorators
        if not isinstance(decorator_list, list):
            decorator_list=[decorator_list]
        for attr in dir(cls):
            if callable(getattr(cls, attr)) and not attr.startswith("_") and attr !="log" and attr !="notify":
                func = getattr(cls,attr)
                for decorator in decorator_list[::-1]: 
                    func = decorator(func)
                setattr(cls, attr, func)
        return cls
    return decorate

#compares two dataframes and tries to convert columns so that match types if they don't otherwise
def compare_data_schemas(obj, data, prev_data):
    #check to see if schemas are equal
    ok = True 
    if not data.equals(prev_data): 
        #if they are not equal, try to set schemas to be the same. 
        ok = False
        try: 
            data = convert_col_types(data, prev_data)
            ok = True
        except Exception as e: 
            obj.notify("Data Comparison failed. New and old dataframes do not have compatible column types. Error: %s" %str(e), "ERROR")
            raise 
    return data, prev_data, True

def convert_col_types(data, prev_data):
    df_a = data.copy()
    df_b = prev_data.copy()
    for col in df_a.columns:
        if col in df_b.columns: #if col is net new, we can skip
            if df_a[col].dtype != df_b[col].dtype:
                df_a[col] = df_a[col].astype(df_b[col].dtype)
    return df_a

#True = lower is better
def get_metric_direction(perf_metric): 
    reverse = False 
    if perf_metric in ["mse", "rmse", "mae", "mape", "smape"]: 
        reverse = True 
    return reverse

#labels should be something like data["y_train"].unique()
def get_multiclass(labels): 
    classification_type = "binary"
    num_classes = len(labels)
    if num_classes > 2: 
        classification_type = "multiclass"

    return classification_type


def create_df_from_file(source_file, engine="pyarrow", **kwargs): 
    _, source_file_type = str(source_file).split("/")[-1].split(".")
    data = pd.DataFrame()
    if source_file_type == "csv":
        data = pd.read_csv(source_file, engine=engine, **kwargs)
    elif source_file_type in ["parquet", "pq"]:
        data = pd.read_parquet(source_file, engine=engine, **kwargs)
    else:
        raise Exception("Unsupported file type: %s" % source_file_type)
    
    return data

def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))




def __get_template_dir__(): 
    import os 

    LOLPOP_DIR = os.path.dirname(os.path.realpath(__file__))
    TEMPLATE_DIR = LOLPOP_DIR + "/templates"

    return TEMPLATE_DIR

__template_path__ = __get_template_dir__()


from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.component.data_connector.local_data_connector import LocalDataConnector
from lolpop.utils import common_utils as utils

from pathlib import Path 


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class LocalDataTransformer(BaseDataTransformer):

    #use load_config to allow setting "DBT_TARGET", "DBT_PROFILE", "DBT_PROJECT_DIR", "DBT_PROFILES_DIR",  via env variables
    __REQUIRED_CONF__ = {
        "config": ["transformer_path"]
    }

    def __init__(self, components={}, *args, **kwargs):
        super().__init__(components=components, *args, **kwargs)

        transformer_path = Path(self._get_config("transformer_path")) 
        transformer_func = self._get_config("transformer_func", "transform")
        
        if transformer_path.exists(): 
            transformer = utils.load_plugin(transformer_path, self)

            self._transform = getattr(transformer, transformer_func)
        else: 
            self.log("Transformer path not found %s" %transformer_path)

        #load data connector
        data_connector_cl_name=self._get_config("data_connector", "LocalDataConnector")
        data_connector_config = self._get_config(
            "data_connector_config", {})
        data_connector_cl = utils.load_class(data_connector_cl_name)
        data_connector = data_connector_cl(
            conf = data_connector_config, pipeline_conf = {}, runner_conf = {}, 
            components={"logger": self.logger})
        self.data_connector = data_connector

    def transform(self, input_data, *args, **kwargs):
        """Runs local transformer workflow in python.

        Args:
            input_data (String): Names of file to load and pass to transformer file.

        Returns:
            pd.DataFrame: the transformed data
        """
        data = self.data_connector.get_data(input_data)
        kwargs = self._get_config("transformer_kwargs",{})

        data_out = self._transform(data, **kwargs)

        return data_out



from lolpop.component.model_visualizer.base_model_visualizer import BaseModelVisualizer
from lolpop.utils import common_utils as utils
from yellowbrick.classifier import ClassificationReport, ConfusionMatrix, ROCAUC, PrecisionRecallCurve, DiscriminationThreshold, ClassPredictionError
from yellowbrick.target import ClassBalance, FeatureCorrelation
import os 
from matplotlib import pyplot as plt
import matplotlib 
matplotlib.use('Agg')

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class YellowbrickModelVisualizer(BaseModelVisualizer): 

    def generate_viz(self, data, model, model_version): 
        
        if self.problem_type == "classification":
            classification_type = utils.get_multiclass(data["y_train"])
            #model = model._get_model() 

            #generate chart for each data split
            #this way is a little inefficient because we have to keep refitting viz
            #but the viz object seems to not allow re-using pyplot. This was the only way 
            #to get all plots to show -- by recreating the object for each split
            for split in set([x.split("_")[1] for x in data.keys()]): 
                #metric report
                viz = ClassificationReport(model)
                self._save_plot(viz, data, split, model_version, "yb_classification_report")

                #confusion matrix
                viz = ConfusionMatrix(model)
                self._save_plot(viz, data, split, model_version, "yb_confusion_matrix")

                #rocauc
                viz = ROCAUC(model)
                self._save_plot(viz, data, split, model_version, "yb_rocauc")

                #comment out for now -- errors if not all classes are represented in the predictions
                ##pr curve
                #viz = PrecisionRecallCurve(model, per_class=True) #per class handles multiclass use cases
                #self._save_plot(viz, data, split, model_version, "yb_prauc")

                #class error
                viz = ClassPredictionError(model)
                self._save_plot(viz, data, split, model_version, "yp_class_error")

                if classification_type == "binary":
                    #discrimination threshold -- only works for binary
                    viz = DiscriminationThreshold(model) 
                    self._save_plot(viz, data, split, model_version, "yb_disctreshold")

            #class balance
            viz = ClassBalance()
            viz.fit(data["y_train"])
            self._save_pyplot("yb_class_balance", "train", model_version)

            #class balance
            viz = FeatureCorrelation(labels=data["X_train"].columns)
            viz.fit(data["X_train"],data["y_train"])
            self._save_pyplot("yb_feature_correlation", "train", model_version)

            #note: could be fun to add some model selection viz here as well: https://www.scikit-yb.org/en/latest/api/model_selection/index.html
            # and yb has some interesting feature analysis viz that we could use in data processing: https://www.scikit-yb.org/en/latest/api/features/index.html

    def _save_plot(self, viz, data, split, model_version, plot_name): 
        viz.fit(data["X_train"], data["y_train"])
        viz.score(data["X_%s" %split], data["y_%s" %split])
        self._save_pyplot(plot_name, split, model_version)

    def _save_pyplot(self, name, label, model_version): 
        local_path = "%s/%s/%s" %(self._get_config("local_dir"), self.metadata_tracker.get_resource_id(model_version), label)
        os.makedirs(local_path, exist_ok=True)
        key = "%s_%s"%(name, label)
        local_file = "%s/%s.png" %(local_path, key)
        plt.savefig(local_file)
        artifact = self.metadata_tracker.log_artifact(model_version, id = key, path = local_file, external=False)
        #plt.clf()
        plt.close()

        return artifact
